@@ -3,14 +3,21 @@ import itertools
 import re
 from enum import Enum
 from typing import Optional, Set, Type, TypeVar, Mapping, Sequence, Union, \
-    ClassVar, Tuple, Generic
+    ClassVar, Tuple, Generic, overload
 
 import attr
 import urllib.parse
 
 
-_segment_regex: re.Pattern = re.compile(r'[a-zA-Z0-9:._+-~]*')
-_segment_allowed_characters: Set[str] = set(':._+-~')
+# according to: https://www.ietf.org/rfc/rfc1738.txt
+# > Thus, only alphanumerics, the special characters "$-_.+!*'(),", and
+# >    reserved characters used for their reserved purposes may be used
+# >    unencoded within a URL.
+# Additionally, tilde (~) is allowed due to it being used for the default COS bucket.
+from ibm_watson_studio_pipelines.client_errors import NoSuchOverloadError
+
+_segment_regex: re.Pattern = re.compile(r'(([a-zA-Z0-9$\-_.+!*\')(~])|(%[0-9a-fA-F]{2}))*')
+_escape_regex: re.Pattern = re.compile(r'%[0-9a-fA-F]{2}')
 
 Self = TypeVar('Self')
 
@@ -19,7 +26,17 @@ def _validate_segment(segment: str) -> None:
         return
 
     forbidden_chars = set()
-    for ch in segment:
+    i = -1
+    while i != len(segment):
+        ch = segment[i]
+        i += 1
+        if ch == '%':
+            # maybe take the next two?
+            if i+2 < len(segment) and _escape_regex.fullmatch(segment[i:i+2]):
+                # it's an escape
+                i += 2
+                continue
+
         if ch in forbidden_chars:
             continue
         if not _segment_regex.fullmatch(ch):
@@ -205,30 +222,25 @@ class Path:
             result += self.file.path
         return result
 
-    @classmethod
-    def from_segments(
-        cls: Type[Self],
-        segments: Sequence[str],
-        has_query: bool,
-    ) -> Self:
-        ns = len(segments)
-        if ns == 0 or (ns == 1 and segments[0] == ""):
-            raise TypeError("At least one path segment expected")
+    @staticmethod
+    def _maybe_parse_scope(
+        segments: Sequence[str]
+    ) -> Tuple[Optional[Segment], Sequence[str]]:
+        assert len(segments) > 0
 
-        scope=None
-        resource=None
-        file=None
+        scope: Optional[Segment] = None
 
         # absolute paths start from "/", so the first segment is empty
         res_idx = 0
         if segments[0] == "":
+            assert len(segments) > 1
             el = segments[1]
             if ScopeType.is_member(el):
                 scope = Segment(type=ScopeType.from_string(el))
             else:
                 raise TypeError(f"Invalid CPD scope type: {segments[1]}")
 
-            if ns > 2:
+            if len(segments) > 2:
                 scope = attr.evolve(scope, id=segments[2])
                 res_idx = 3
             else:
@@ -239,66 +251,101 @@ class Path:
         if res_idx is not None:
             remaining_segments = segments[res_idx:]
 
+        return scope, remaining_segments
+
+    @staticmethod
+    def _parse_resource_and_file(
+        segments: Sequence[str],
+        has_scope: bool,
+        has_query: bool,
+    ) -> Tuple[Optional[Segment], Optional[File]]:
+        ns = len(segments)
+
+        resource: Optional[Segment] = None
+        file: Optional[File] = None
+
+        # files/...
+        if ns > 1 and segments[0] == SpecialSegmentsType.FILES.value:
+            file = File(path="/".join(segments[1:]))
+        # asset-123
+        else:
+            if ns == 1 and not has_query:
+                el = segments[0]
+                if has_scope:
+                    # cannot: /spaces/abc/asset-123
+                    if ResourceType.is_member(el) or SpecialAssetsType.is_member(el):
+                        # /spaces/abc/deployments
+                        # lacks query?
+                        raise TypeError(f"CPD resource path to '{el}' lacks either id or query.")
+                    else:
+                        # /spaces/abc/asset-123
+                        # lacks `assets` mid-path?
+                        raise TypeError(f"CPD path is absolute but looks like relative. Maybe lacking 'assets' before id.")
+
+                # note: this is one of the few cases in which the path will look different
+                # asset-123 --> assets/asset-123
+                resource = Segment(type=ResourceType.ASSETS, id=el)
+            else: # ns > 1 or has_query
+                el = segments[0]
+                # deployments/abc, deployments?name=abc
+                if ResourceType.is_member(el):
+                    resource = Segment(type=ResourceType.from_string(el))
+                # notebooks/abc, notebooks?name=abc
+                elif SpecialAssetsType.is_member(el):
+                    resource = Segment(type=SpecialAssetsType.from_string(el))
+                # files/...
+                elif el == SpecialSegmentsType.FILES.value:
+                    file = File(path="/".join(segments[1:]))
+                    # rest of the path is file path, do not consider
+                    ns = 1
+                else:
+                    raise TypeError(f"Invalid CPD resource type: {segments[0]}")
+
+                # deployments/abc
+                if ns > 1:
+                    assert resource is not None # as seen above
+                    resource = attr.evolve(resource, id=segments[1])
+
+            if ns > 2:
+                # connections/_/files/~/...
+                if segments[2] == SpecialSegmentsType.FILES.value:
+                    if resource is None or resource.type != SpecialAssetsType.CONNECTIONS:
+                        if resource is not None:
+                            typ = resource.type
+                            raise TypeError(f"File path only allowed for connection asset, but used on: {typ}")
+                        else:
+                            raise TypeError(f"File path only allowed for connection asset")
+                    file = File(path="/".join(segments[3:]))
+                else:
+                    raise TypeError(f"Unexpected path segments: {'/'.join(segments[2:])}")
+        return resource, file
+
+    @classmethod
+    def from_segments(
+        cls: Type[Self],
+        segments: Sequence[str],
+        has_query: bool,
+    ) -> Self:
+        ns = len(segments)
+        if ns == 0 or (ns == 1 and segments[0] == ""):
+            raise TypeError("At least one path segment expected")
+
+        scope, remaining_segments = cls._maybe_parse_scope(segments)
+
         # if no remaining segments, finish
         if not remaining_segments:
             return Path(
                 scope=scope,
-                resource=resource,
-                file=file,
+                resource=None,
+                file=None,
             )
 
-        segments = remaining_segments
-        ns = len(segments)
-        # asset-123
-        if ns == 1 and not  has_query:
-            el = segments[0]
-            if scope is not None:
-                # cannot: /spaces/abc/asset-123
-                if ResourceType.is_member(el) or SpecialAssetsType.is_member(el):
-                    # /spaces/abc/deployments
-                    # lacks query?
-                    raise TypeError(f"CPD resource path to '{el}' lacks either id or query.")
-                else:
-                    # /spaces/abc/asset-123
-                    # lacks `assets` mid-path?
-                    raise TypeError(f"CPD path is absolute but looks like relative. Maybe lacking 'assets' before id.")
-
-            # note: this is one of the few cases in which the path will look different
-            # asset-123 --> assets/asset-123
-            resource = Segment(type=ResourceType.ASSETS, id=el)
-        else: # ns > 1 or has_query
-            el = segments[0]
-            # deployments/abc, deployments?name=abc
-            if ResourceType.is_member(el):
-                resource = Segment(type=ResourceType.from_string(el))
-            # notebooks/abc, notebooks?name=abc
-            elif SpecialAssetsType.is_member(el):
-                resource = Segment(type=SpecialAssetsType.from_string(el))
-            # files/...
-            elif el == SpecialSegmentsType.FILES.value:
-                file = File(path="/".join(segments[1:]))
-                # rest of the path is file path, do not consider
-                ns = 1
-            else:
-                raise TypeError(f"Invalid CPD resource type: {segments[0]}")
-
-            # deployments/abc
-            if ns > 1:
-                assert resource is not None # as seen above
-                resource = attr.evolve(resource, id=segments[1])
-
-        if ns > 2:
-            # connections/files/~/...
-            if segments[2] == SpecialSegmentsType.FILES.value:
-                if resource is None or resource.type != SpecialAssetsType.CONNECTIONS:
-                    if resource is not None:
-                        typ = resource.type
-                        raise TypeError(f"File path only allowed for connection asset, but used on: {typ}")
-                    else:
-                        raise TypeError(f"File path only allowed for connection asset")
-                file = File(path="/".join(segments[3:]))
-            else:
-                raise TypeError(f"Unexpected path segments: {'/'.join(segments[2:])}")
+        has_scope = scope is not None
+        resource, file = cls._parse_resource_and_file(
+            remaining_segments,
+            has_scope,
+            has_query,
+        )
 
         return Path(
             scope=scope,
@@ -426,21 +473,19 @@ class CpdPath:
         # creates different kinds of CpdPath depending on path
         data = CpdPathData.from_string(path)
 
-        if data.path.scope is not None and data.path.resource is None:
-            _validate_scope_query(data)
-            if data.path.file is None:
-                # scope
-                return CpdScope(data)
+        if data.path.file is not None:
+            if data.path.resource is not None:
+                # connection file
+                return CpdConnectionFile(data)
             else:
                 # scope file
                 return CpdScopeFile(data)
         elif data.path.resource is not None:
-            if data.path.file is None:
-                # resource
-                return CpdResource(data)
-            else:
-                # connection file
-                return CpdConnectionFile(data)
+            # resource
+            return CpdResource(data)
+        elif data.path.scope is not None:
+            # scope
+            return CpdScope(data)
         else:
             raise TypeError(f"Uknown kind of CpdPath: {str(data)}")
 
@@ -509,6 +554,8 @@ class CpdPath:
         query_str = ""
         if query_obj:
             query_str = str(query_obj)
+        if query_str == "":
+            query_str = None
         return query_str
 
 class CpdPathImpl(CpdPath):
@@ -527,7 +574,31 @@ class CpdScope(CpdPathImpl):
     def scope_type(self) -> str:
         return self.data.path.scope.type.value
 
-class CpdScopeFile(CpdScope):
+    def file(self, path: str) -> 'CpdScopeFile':
+        """
+        Creates a path to a file in the given scope.
+
+        The leading slash is optional.
+
+        Example::
+
+            scope = CpdScope.from_string("cpd://dev/projects/123456789")
+            file_path_0 = scope.file("/files/abc/def")
+            file_path_1 = scope.file("files/abc/def")
+            assert file_path_0 == file_path_1
+            assert file_path_0.file() == "/files/abc/def"
+        """
+        if self.query() is not None:
+            raise RuntimeError(f".file(...) is only available for non-query paths")
+
+        new_file_path = path.lstrip('/')
+        new_file = File(path=new_file_path)
+        new_path = attr.evolve(self.data.path, file=new_file)
+        new_data = attr.evolve(self.data, path=new_path)
+        new_cpd_path = CpdScopeFile(data=new_data)
+        return new_cpd_path
+
+class CpdScopeFile(CpdPathImpl):
     def file_path(self) -> str:
         """
         Extract the file path.
@@ -595,10 +666,114 @@ class CpdResource(CpdPathImpl):
         """
         return self.data.path.resource.id
 
-class CpdConnectionFile(CpdResource):
+    @overload
+    def file(self, path: str) -> 'CpdConnectionFile': ...
+    @overload
+    def file(self, bucket_name: str, path: str) -> 'CpdConnectionFile': ...
+
+    def file(self, *args, **kwargs) -> 'CpdConnectionFile':
+        """
+        Creates a path to a file in the given connection.
+
+        The leading slash is optional.
+
+        Will fail if the resource is not a connection.
+
+        Example::
+
+            scope = CpdResource.from_string("cpd://dev/projects/123456789/connections/3141592")
+            file_path_0 = scope.file("/files/abc/def")
+            file_path_1 = scope.file("~", "/files/abc/def")
+            assert file_path_0 == file_path_1
+            assert file_path_0.bucket_name() == "~"
+            assert file_path_0.file() == "/files/abc/def"
+        """
+        if len(args) == 2:
+            bucket_name = args[0]
+            path = args[1]
+        elif len(args) == 1:
+            if 'path' in kwargs:
+                bucket_name = args[0]
+                path = kwargs['path']
+            else:
+                bucket_name = '~'
+                path = args[0]
+        elif 'path' in kwargs and 'bucket_name' in kwargs:
+            bucket_name = kwargs['bucket_name']
+            path = kwargs['path']
+        elif 'path' in kwargs:
+            bucket_name = '~'
+            path = kwargs['path']
+        else:
+            raise NoSuchOverloadError(f'{type(self).__name__}.file(...)', args, kwargs)
+
+        if self.query() is not None:
+            raise RuntimeError(f".file(...) is only available for non-query paths")
+
+        if self.resource_type() != "connections":
+            raise RuntimeError(f".file(...) is only available for connection resource, but used on '{self.resource_type()}'")
+
+        bucket_name = bucket_name.strip('/')
+        path = path.lstrip('/')
+        new_file_path = f'{bucket_name}/{path}'
+
+        new_file = File(path=new_file_path)
+        new_path = attr.evolve(self.data.path, file=new_file)
+        new_data = attr.evolve(self.data, path=new_path)
+        new_cpd_path = CpdConnectionFile(data=new_data)
+        return new_cpd_path
+
+class CpdConnectionFile(CpdPathImpl):
+    def scope(self) -> Optional[CpdScope]:
+        """
+        Extract the scope path, if any.
+
+        Example::
+
+            scope_path = CpdPath.from_string("cpd://dev/projects/123456789")
+            resource_path = CpdPath.from_string("cpd://dev/projects/123456789/connections/3141592")
+            assert resource_path.scope() == scope_path
+        """
+
+        if self.data.path.scope is None:
+            return None
+        return CpdScope(
+            data=CpdPathData(
+                has_prefix=self.data.has_prefix,
+                ctx=self.data.ctx,
+                path=Path(
+                    scope=self.data.path.scope,
+                    resource=None,
+                    file=None,
+                ),
+                query=None,
+            )
+        )
+
+    def resource_type(self) -> str:
+        """
+        Return the resource type.
+
+        Example::
+
+            path = CpdPath.from_string("cpd://dev/projects/123456789/connections/3141592")
+            assert path.resource_type() == "connections"
+        """
+        return self.data.path.resource.type.value
+
     def resource_id(self) -> str:
-        # always present due to syntax
-        return super().resource_id()
+        """
+        Return the resource ID, if any.
+
+        Example::
+
+            path = CpdPath.from_string("cpd://dev/projects/123456789/connections/3141592")
+            assert path.resource_id() == "3141592"
+
+            path = CpdPath.from_string("cpd://dev/projects/123456789/connections?name=foobar")
+            assert path.resource_id() is None
+        """
+        return self.data.path.resource.id
 
     def bucket_name(self) -> str:
         """
